@@ -7,10 +7,9 @@ export KO_DOCKER_REPO=ko.local
 
 set -u
 
-(
-  cd ../.. && \
-  ko build --platform=all --base-import-paths --tags="latest" --image-user=1000 --image-annotation="" --image-label="" ./cmd/anubis -L
-)
+source ../lib/lib.sh
+
+build_anubis_ko
 
 rm -rf ./var/repos ./var/foo
 mkdir -p ./var/repos
@@ -34,4 +33,4 @@ sleep 2
   git push -u http://localhost:3000/git/foo.git master
 )
 
-docker compose down
\ No newline at end of file
+exit 0
\ No newline at end of file
