@@ -3,6 +3,7 @@ package web
 import (
 	"github.com/a-h/templ"
 
+	"github.com/TecharoHQ/anubis/lib/challenge"
 	"github.com/TecharoHQ/anubis/lib/localization"
 	"github.com/TecharoHQ/anubis/lib/policy/config"
 )
@@ -11,18 +12,14 @@ func Base(title string, body templ.Component, impressum *config.Impressum, local
 	return base(title, body, impressum, nil, nil, localizer)
 }
 
-func BaseWithChallengeAndOGTags(title string, body templ.Component, impressum *config.Impressum, challenge string, rules *config.ChallengeRules, ogTags map[string]string, localizer *localization.SimpleLocalizer) (templ.Component, error) {
+func BaseWithChallengeAndOGTags(title string, body templ.Component, impressum *config.Impressum, challenge *challenge.Challenge, rules *config.ChallengeRules, ogTags map[string]string, localizer *localization.SimpleLocalizer) templ.Component {
 	return base(title, body, impressum, struct {
 		Rules     *config.ChallengeRules `json:"rules"`
-		Challenge string                 `json:"challenge"`
+		Challenge any                    `json:"challenge"`
 	}{
 		Challenge: challenge,
 		Rules:     rules,
-	}, ogTags, localizer), nil
-}
-
-func Index(localizer *localization.SimpleLocalizer) templ.Component {
-	return index(localizer)
+	}, ogTags, localizer)
 }
 
 func ErrorPage(msg, mail string, localizer *localization.SimpleLocalizer) templ.Component {
